/* This file was automatically generated, do not edit */
#include "common.h"
extern const unsigned p256_n_tables;
extern const unsigned p256_window_size;
extern const unsigned p256_points_per_table;
extern const uint64_t p256_tables[52][32][2][4];

